from z3 import *

def SHF(data):
	res = 0x2FD2B4
	for d in data:
		res ^= int(d)
		res *= 0x66EC73
		res &= 0xffffffffffffffff
	return res


def sanity():

	b = bytearray([0x89, 0x50, 0x4E, 0x47])
	res = SHF(b)
	assert(res == 0x9141A25EC72A7B32)

sanity()

def trySolution(values):

	#values = [7, 54, 61, 61, 59, 56, 58, 80, 83, 79, 85, 83, 38, 12]

	# read output
	data = bytearray(open("output.png.orig", "rb").read())
	chunks = []

	for i in range(0, 14):
		
		chunk = data[0x1000 * i:0x1000 * (i+1)]
		# put values in place
		chunk[10] = values[i]
		chunks.append(chunk)

	lastChunk = data[0x1000 * 14:]
	chunks.append(lastChunk)
	#print("[~] last chunk's size = %d / %x" % (len(lastChunk), len(lastChunk)))

	# check flag
	bufF000 = bytearray(0xe000 + len(lastChunk))
	for i in range(len(bufF000)):
		bufF000[i] = chunks[int(i / 0x1000)][i % 0x1000]

	shf = SHF(bufF000)

	flag = bytearray(4)
	flag[0] = ((shf & 0xff) - 0x7d) & 0xff
	flag[1] = (((shf & 0xff00) >> 8) + 0x7c) & 0xff
	flag[2] = (((shf & 0xff0000) >> 16)) & 0xff
	flag[3] = (((shf & 0xff000000) >> 24) - 0x51) & 0xff

	#print(flag)

	# reorder
	for k in range(7):
		if not ( (chunks[2 * k][0] + chunks[2 * k +1][0]) & 1 ):
			#print("[+] swapping %d" % k)
			chunks[2 * k], chunks[2 * k + 1] =  chunks[2 * k + 1], chunks[2 * k]

	name = "_".join(["%x" % x for x in flag])
	blob = bytearray()
	for c in chunks:
		blob += c
	open("out/data_%s" % name, "wb").write(blob)

	if flag == bytearray("Flag".encode("utf-8")):
		print("[+] done!")
		exit()

def pow3_64(v):
	return v*v*v

V_BITS = 8

done = []

while(1):

	vals = [BitVec('v%d' % i, V_BITS) for i in range(0, 14)] 

	s = Solver()

	s.add(vals[0] == 7)
	s.add(vals[13] == 0xc)

	for i in range(1, 7):
		s.add(vals[i] - 52 <= 9)
		s.add(vals[i] - 52 >= 0)

	for i in range(7, 12):
		s.add(vals[i] - 77 <= 9)
		s.add(vals[i] - 77 >= 0)

	s.add(vals[12] - 34 <= 9)
	s.add(vals[12] - 34 >= 0)

	v = vals
	s.add( (pow3_64(v[1]) + pow3_64(v[2]) + pow3_64(v[3])) & 0xff == 0x62 )
	s.add( (pow3_64(v[4]) + pow3_64(v[5]) + pow3_64(v[6]) + pow3_64(v[7])) & 0xff == 0x6b )
	s.add( (pow3_64(v[9]) + pow3_64(v[10]) + pow3_64(v[11]) + pow3_64(v[12])) & 0xff == 0xbf )

	for d in done:
		conds = []
		for i, dd in enumerate(d):
			conds.append(dd == v[i])
		s.add(Not(And(conds)))

	if s.check() != sat:
		print(":(")
		exit()

	m = s.model()
	#print(m)
	res = [m[vv].as_long() for vv in v]
	print(res)
	#exit()
	done.append(res)
	
	trySolution(res)
